from subprocess import CalledProcessError
from ..helpers import B2LuigiTestCase


class TestIgnoreAdditionalCommandLineArgs(B2LuigiTestCase):
    def test_accept_user_args_if_ignored_by_b2luigi(self):
        """
        Check that we can use our own CLI args in scripts if
        ``ignore_additional_command_line_args`` is ``True``, which means they
        should be ignore by ``b2luigi.process``.
        """
        self.call_file(
            "cli/process_with_user_args_ignored_by_b2luigi.py",
            cli_args=["--hello", "world", "1", "2"],
        )

    def test_user_args_when_not_ignored_raises_error(self):
        """
        Check that we can't use undefined CLI args in scripts if
        ``ignore_additional_command_line_args=False`` in ``b2luigi.process.
        """
        with self.assertRaises(CalledProcessError):
            self.call_file(
                "cli/process_with_user_args_not_ignored_by_b2luigi.py",
                cli_args=["--hello", "world"],
            )



import json
import re
import subprocess
import os

from b2luigi.batch.processes import BatchProcess, JobStatus
from b2luigi.batch.cache import BatchJobStatusCache
from b2luigi.core.utils import get_log_file_dir
from b2luigi.core.executable import create_executable_wrapper
from b2luigi.core.settings import get_setting


class LSFJobStatusCache(BatchJobStatusCache):
    def _ask_for_job_status(self, job_id=None):
        if job_id:
            output = subprocess.check_output(["bjobs", "-json", "-o", "jobid stat", str(job_id)])
        else:
            output = subprocess.check_output(["bjobs", "-json", "-o", "jobid stat"])
        output = output.decode()
        output = json.loads(output)["RECORDS"]

        for record in output:
            self[record["JOBID"]] = record["STAT"]


_batch_job_status_cache = LSFJobStatusCache()


class LSFProcess(BatchProcess):
    """
    Reference implementation of the batch process for a LSF batch system.

    Additional to the basic batch setup (see :ref:`batch-label`), there are
    LSF-specific :meth:`settings <b2luigi.set_setting>`. These are:

    * the LSF queue: ``queue``.
    * the LSF job name: ``job_name``.

    For example:

      .. code-block:: python

        class MyLongTask(b2luigi.Task):
            queue = "l"
            job_name = "my_long_task"

      The default queue is the short queue ``"s"``. If no ``job_name`` is set the task
      will appear as ::

        <result_dir>/parameter1=value/.../executable_wrapper.sh"

      when running ``bjobs``.

    * By default, the environment variables from the scheduler are copied to
      the workers.
      This also applies we start in the same working directory and can reuse
      the same executable etc.
      Normally, you do not need to supply ``env_script`` or alike.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_job_id = None

    def get_job_status(self):
        if not self._batch_job_id:
            return JobStatus.aborted

        try:
            job_status = _batch_job_status_cache[self._batch_job_id]
        except KeyError:
            return JobStatus.aborted

        if job_status == "DONE":
            return JobStatus.successful
        if job_status == "EXIT":
            return JobStatus.aborted

        return JobStatus.running

    def start_job(self):
        command = ["bsub", "-env all"]

        queue = get_setting("queue", task=self.task, default=False)
        if queue is not False:
            command += ["-q", queue]

        job_name = get_setting("job_name", task=self.task, default=False)
        if job_name is not False:
            command += ["-J", job_name]

        log_file_dir = get_log_file_dir(self.task)
        os.makedirs(log_file_dir, exist_ok=True)

        stdout_log_file = os.path.join(log_file_dir, "stdout")
        stderr_log_file = os.path.join(log_file_dir, "stderr")

        command += ["-eo", stderr_log_file, "-oo", stdout_log_file]

        executable_file = create_executable_wrapper(self.task)
        command.append(executable_file)

        output = subprocess.check_output(command)
        output = output.decode()

        # Output of the form Job <72065926> is submitted to default queue <s>.
        match = re.search(r"<[0-9]+>", output)
        if not match:
            raise RuntimeError("Batch submission failed with output " + output)

        self._batch_job_id = match.group(0)[1:-1]

    def terminate_job(self):
        if not self._batch_job_id:
            return

        subprocess.run(["bkill", self._batch_job_id], stdout=subprocess.DEVNULL, check=False)



import json
import os
import warnings

from ..helpers import B2LuigiTestCase

import b2luigi
from b2luigi.core.settings import DeprecatedSettingsWarning


class TaskTestCase(B2LuigiTestCase):
    def setUp(self):
        super().setUp()

        b2luigi.clear_setting("my_setting")

    def test_set_by_function(self):
        self.assertRaises(ValueError, b2luigi.get_setting, "my_setting")

        self.assertEqual("default", b2luigi.get_setting("my_setting", "default"))

        b2luigi.set_setting("my_setting", "my value")

        self.assertEqual("my value", b2luigi.get_setting("my_setting"))
        self.assertEqual("my value", b2luigi.get_setting("my_setting", "default"))

    def test_set_by_file(self):
        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my file value"}, f)

        self.assertEqual("my file value", b2luigi.get_setting("my_setting"))

        b2luigi.set_setting("my_setting", "my value")

        self.assertEqual("my value", b2luigi.get_setting("my_setting"))

    def test_set_by_parent_file(self):
        os.mkdir("child_path")
        os.chdir("child_path")

        with open("../settings.json", "w") as f:
            json.dump({"my_setting": "my parent value"}, f)

        self.assertEqual("my parent value", b2luigi.get_setting("my_setting"))

        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my child value"}, f)

        self.assertEqual("my child value", b2luigi.get_setting("my_setting"))

    def test_set_by_task_or_file(self):
        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my file value"}, f)

        b2luigi.set_setting("my_second_setting", "my value")

        task = b2luigi.Task()
        setattr(task, "my_third_setting", "my task value")

        self.assertEqual("my file value", b2luigi.get_setting("my_setting", task=task))
        self.assertEqual("my value", b2luigi.get_setting("my_second_setting", task=task))
        self.assertEqual("my task value", b2luigi.get_setting("my_third_setting", task=task))

    def test_deprecated_settings(self):
        self.assertRaises(ValueError, b2luigi.get_setting, key="my_setting", deprecated_keys=["my_old_setting"])

        b2luigi.set_setting("my_old_setting", "my value")

        with warnings.catch_warnings(record=True) as w:
            self.assertEqual("my value", b2luigi.get_setting("my_setting", deprecated_keys=["my_old_setting"]))

            self.assertEqual(len(w), 1)
            self.assertIsInstance(w[-1].message, DeprecatedSettingsWarning)
            self.assertIn("deprecated", str(w[-1].message))

        b2luigi.set_setting("my_setting", "my new_value")

        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(
                "my new_value", b2luigi.get_setting("my_setting", default="default", deprecated_keys=["my_old_setting"])
            )

            self.assertEqual(len(w), 0)



import subprocess

from ..helpers import B2LuigiTestCase

import os


class DispatchTaskTestCase(B2LuigiTestCase):
    def test_failing_task_class(self):
        for file_name in ["core/dispatch_1.py", "core/dispatch_2.py"]:
            out = self.call_file(file_name, stderr=subprocess.STDOUT)

            self.assertTrue(os.path.exists("results/some_file.txt"))
            self.assertFalse(os.path.exists("results/some_other_file.txt"))
            self.assertTrue(os.path.exists("logs/MyTask/stderr"))
            self.assertTrue(os.path.exists("logs/MyTask/stdout"))

            with open("logs/MyTask/stdout", "r") as f:
                stdout_content = f.readlines()
                self.assertIn("Hello!\n", stdout_content)
                self.assertIn("Bye!\n", stdout_content)

            self.assertIn(b"Task MyTask failed!", out.splitlines())
            self.assertIn(b"Please have a look into the log files in", out.splitlines())
            self.assertIn(b"RuntimeError: Execution failed with return code -11", out.splitlines())

    def test_env_and_script_dispatch(self):
        for file_name in ["core/dispatch_with_env_and_script.py"]:
            out = self.call_file(file_name, stderr=subprocess.STDOUT)

            self.assertTrue(os.path.exists("logs/MyTask/stderr"))
            self.assertTrue(os.path.exists("logs/MyTask/stdout"))

            with open("logs/MyTask/stdout", "r") as f:
                stdout_content = f.readlines()
                self.assertIn("MY_SECRET_VARIABLE 42\n", stdout_content)
                self.assertIn("MY_SECOND_SECRET_VARIABLE 47\n", stdout_content)

            self.assertIn(
                b"This progress looks :) because there were no failed tasks or missing dependencies", out.splitlines()
            )



from b2luigi.core import utils

import luigi

from b2luigi.core.utils import create_output_file_name


class Task(luigi.Task):
    """
    Drop in replacement for ``luigi.Task`` which is 100% API compatible.
    It just adds some useful methods for handling output file name generation using
    the parameters of the task.
    See :ref:`quick-start-label` on information on how to use the methods.

    Example:

        .. code-block:: python

          class MyAverageTask(b2luigi.Task):
              def requires(self):
                  for i in range(100):
                      yield self.clone(MyNumberTask, some_parameter=i)

              def output(self):
                  yield self.add_to_output("average.txt")

              def run(self):
                  # Build the mean
                  summed_numbers = 0
                  counter = 0
                  for input_file in self.get_input_file_names("output_file.txt"):
                      with open(input_file, "r") as f:
                          summed_numbers += float(f.read())
                          counter += 1

                  average = summed_numbers / counter

                  with self.get_output_file("average.txt").open("w") as f:
                      f.write(f"{average}\\n")
    """

    def add_to_output(self, output_file_name):
        """
        Call this in your ``output()`` function to add a target to the list of files,
        this task will output.
        Always use in combination with ``yield``.
        This function will automatically add all current parameter values to
        the file name when used in the form::

            <result-path>/param1=value1/param2=value2/.../<output-file-name.ext>

        This function will automatically use a ``LocalTarget``.
        If you do not want this, you can override the :obj:`_get_output_file_target` function.

        Example:
            This adds two files called ``some_file.txt`` and ``some_other_file.txt`` to the output:

            .. code-block:: python

              def output(self):
                  yield self.add_to_output("some_file.txt")
                  yield self.add_to_output("some_other_file.txt")

        Args:
            output_file_name (:obj:`str`): the file name of the output file.
                Refer to this file name as a key when using :obj:`get_input_file_names`,
                :obj:`get_output_file_names` or :obj:`get_output_file`.
        """
        return {output_file_name: self._get_output_file_target(output_file_name)}

    @staticmethod
    def _transform_input(input_generator, key=None):
        input_list = utils.flatten_to_list_of_dicts(input_generator)
        file_paths = utils.flatten_to_file_paths(input_list)

        if key is not None:
            return file_paths[key]
        return file_paths

    def get_all_input_file_names(self):
        """
        Return all file paths required by this task.

        Example:

            .. code-block:: python

              class TheSuperFancyTask(b2luigi.Task):
                  def dry_run(self):
                      for name in self.get_all_output_file_names():
                          print(f"\t\toutput:\t{name}")

        """
        for file_names in self._transform_input(self.input()).values():
            for file_name in file_names:
                yield file_name

    def get_input_file_names(self, key=None):
        """
        Get a dictionary of input file names of the tasks, which are defined in our requirements.
        Either use the key argument or dictionary indexing with the key given to :obj:`add_to_output`
        to get back a list (!) of file paths.

        Args:
            key (:obj:`str`, optional): If given, only return a list of file paths with this given key.

        Return:
            If key is none, returns a dictionary of keys to list of file paths.
            Else, returns only the list of file paths for this given key.
        """
        return self._transform_input(self.input(), key)

    def get_input_file_names_from_dict(self, requirement_key, key=None):
        """
        Get a dictionary of input file names of the tasks, which are defined in our requirements.

        The requirement method should return a dict whose values are generator expressions (!)
        yielding required task objects.

        Example:

            .. code-block:: python

              class TaskB(luigi.Task):

                  def requires(self):
                      return {
                          "a": (TaskA(5.0, i) for i in range(100)),
                          "b": (TaskA(1.0, i) for i in range(100)),
                      }

                  def run(self):
                      result_a = do_something_with_a(
                          self.get_input_file_names_from_dict("a")
                      )
                      result_b = do_something_with_b(
                          self.get_input_file_names_from_dict("b")
                      )

                      combine_a_and_b(
                          result_a,
                          result_b,
                          self.get_output_file_name("combined_results.txt")
                      )

                  def output(self):
                      yield self.add_to_output("combined_results.txt")

        Either use the key argument or dictionary indexing with the key given to :obj:`add_to_output`
        to get back a list (!) of file paths.

        Args:
            requirement_key (:obj:`str`): Specifies the required task expression.
            key (:obj:`str`, optional): If given, only return a list of file paths with this given key.

        Return:
            If key is none, returns a dictionary of keys to list of file paths.
            Else, returns only the list of file paths for this given key.
        """
        return self._transform_input(self.input()[requirement_key], key)

    @staticmethod
    def _transform_output(output_generator, key=None):
        output_list = utils.flatten_to_list_of_dicts(output_generator)
        file_paths = utils.flatten_to_file_paths(output_list)

        if key is not None:
            return file_paths[key]
        return file_paths

    def get_all_output_file_names(self):
        """
        Return all file paths created by this task.

        Example:

            .. code-block:: python

              class TheSuperFancyTask(b2luigi.Task):
                  def dry_run(self):
                      for name in self.get_all_output_file_names():
                          print(f"\t\toutput:\t{name}")
        """
        for file_names in self._transform_output(self.output()).values():
            for file_name in file_names:
                yield file_name

    def get_output_file_name(self, key):
        """
        Analogous to :obj:`get_input_file_names` this function returns
        a an output file defined in out output function with
        the given key.

        In contrast to :obj:`get_input_file_names`, only a single file name
        will be returned (as there can only be a single output file with a given name).

        Args:
            key (:obj:`str`): Return the file path with this given key.

        Return:
            Returns only the file path for this given key.
        """
        target = self._get_output_target(key)
        file_paths = utils.flatten_to_file_paths(target)

        return file_paths

    def _get_input_targets(self, key):
        """Shortcut to get the input targets for a given key. Will return a luigi target."""
        input_dict = utils.flatten_to_list_of_dicts(self.input())
        return input_dict[key]

    def _get_output_target(self, key):
        """Shortcut to get the output target for a given key. Will return a luigi target."""
        output_dict = utils.flatten_to_dict(self.output())
        return output_dict[key]

    def _get_output_file_target(self, base_filename, **kwargs):
        file_name = create_output_file_name(self, base_filename, **kwargs)
        return luigi.LocalTarget(file_name)


class ExternalTask(Task, luigi.ExternalTask):
    """Direct copy of :obj:`luigi.ExternalTask`, but with the capabilities of :obj:`Task` added."""

    pass


class WrapperTask(Task, luigi.WrapperTask):
    """Direct copy of :obj:`luigi.WrapperTask`, but with the capabilities of :obj:`Task` added."""

    pass


class NotCompletedTask(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_complete = True

    def complete(self):
        """Custom complete function checking also the child tasks until a check_complete = False is reached"""
        if not super().complete():
            return False

        if not self.check_complete:
            return True

        requires = self.requires()

        try:
            if not requires.complete():
                return False
        except AttributeError:
            for task in requires:
                if not task.complete():
                    return False

        return True



import b2luigi
from b2luigi.basf2_helper.tasks import Basf2PathTask

import example_mdst_analysis


class AnalysisTask(Basf2PathTask):
    # set the batch_system property to use the gbasf2 wrapper batch process for this task
    batch_system = "gbasf2"
    # Must define a prefix for the gbasf2 project name to submit to the grid.
    # b2luigi will then add a hash derived from the luigi parameters to create a unique project name.
    gbasf2_project_name_prefix = b2luigi.Parameter()
    gbasf2_input_dataset = b2luigi.Parameter(hashed=True)
    # Example luigi cut parameter to facilitate starting multiple projects for different cut values
    mbc_lower_cut = b2luigi.IntParameter()

    def create_path(self):
        mbc_range = (self.mbc_lower_cut, 5.3)
        return example_mdst_analysis.create_analysis_path(
            d_ntuple_filename="D_ntuple.root", b_ntuple_filename="B_ntuple.root", mbc_range=mbc_range
        )

    def output(self):
        yield self.add_to_output("D_ntuple.root")
        yield self.add_to_output("B_ntuple.root")


class AnalysisWrapperTask(b2luigi.WrapperTask):
    """
    We use the AnalysisWrapperTask to be able to require multiple analyse tasks with
    different input datasets and cut values. For each parameter combination, a
    different gbasf2 project will be submitted.
    """

    def requires(self):
        input_dataset = (
            "/belle/MC/release-04-01-04/DB00000774/SkimM13ax1/prod00011778/e1003/4S/r00000/mixed/11180100/udst/sub00/"
            "udst_000006_prod00011778_task10020000006.root"
        )
        # if you want to iterate over different cuts, just add more values to this list
        mbc_lower_cuts = [5.15, 5.2]
        for mbc_lower_cut in mbc_lower_cuts:
            yield AnalysisTask(
                mbc_lower_cut=mbc_lower_cut,
                gbasf2_project_name_prefix="luigiExample",
                gbasf2_input_dataset=input_dataset,
                max_event=100,
            )


if __name__ == "__main__":
    main_task_instance = AnalysisWrapperTask()
    n_gbasf2_tasks = len(list(main_task_instance.requires()))
    b2luigi.process(main_task_instance, workers=n_gbasf2_tasks)



import contextlib
import importlib

import itertools
import os
import collections
import sys
import types

import colorama

from b2luigi.core.settings import get_setting


@contextlib.contextmanager
def remember_cwd():
    """Helper contextmanager to stay in the same cwd"""
    old_cwd = os.getcwd()
    try:
        yield
    finally:
        os.chdir(old_cwd)


def product_dict(**kwargs):
    """
    Cross-product the given parameters and return a list of dictionaries.

    Example:

        .. code-block:: python

            >>> list(product_dict(arg_1=[1, 2], arg_2=[3, 4]))
            [{'arg_1': 1, 'arg_2': 3}, {'arg_1': 1, 'arg_2': 4}, {'arg_1': 2, 'arg_2': 3}, {'arg_1': 2, 'arg_2': 4}]

        The thus produced list can directly be used as inputs for a required tasks:

        .. code-block:: python

            def requires(self):
                for args in product_dict(arg_1=[1, 2], arg_2=[3, 4]):
                    yield some_task(**args)

    Parameters:
        kwargs: Each keyword argument should be an iterable

    Return:
        A list of kwargs where each list of input keyword arguments is cross-multiplied with every other.
    """
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


def fill_kwargs_with_lists(**kwargs):
    """
    Return the kwargs with each value mapped to [value] if not a list already.

    Example:
    .. code-block:: python

        >>> fill_kwargs_with_lists(arg_1=[1, 2], arg_2=3)
        {'arg_1': [1, 2], 'arg_2': [3]}

    :param kwargs: The input keyword arguments
    :return: Same as kwargs, but each value mapped to a list if not a list already
    """
    return_kwargs = {}
    for key, value in kwargs.items():
        if value is None:
            value = []
        if isinstance(value, str) or not isinstance(value, collections.Iterable):
            value = [value]
        return_kwargs[key] = value

    return return_kwargs


def flatten_to_file_paths(inputs):
    """
    Take in a structure of something and replace each luigi target by its corresponding path.
    For dicts, it will replace the value as well as the key. The key will however only by the basename of the path.

    :param inputs: A dict or a luigi target
    :return: A dict with the keys replaced by the basename of the targets and the values by the full path
    """
    if not inputs:
        return None

    try:
        return inputs.path
    except AttributeError:
        pass

    if isinstance(inputs, dict):
        return {
            os.path.basename(flatten_to_file_paths(key)): flatten_to_file_paths(value) for key, value in inputs.items()
        }
    if isinstance(inputs, list):
        return [flatten_to_file_paths(value) for value in inputs]
    return inputs


def flatten_to_dict(inputs):
    """
    Return a whatever input structure into a dictionary.
    If it is a dict already, return this.
    If is is an iterable of dict or dict-like objects, return the merged dictionary.
    All non-dict values will be turned into a dictionary with value -> {value: value}

    Example:
    .. code-block:: python

        >>> flatten_to_dict([{"a": 1, "b": 2}, {"c": 3}, "d"])
        {'a': 1, 'b': 2, 'c': 3, 'd': 'd'}

    :param inputs: The input structure
    :return: A dict constructed as described above.
    """
    inputs = _flatten(inputs)
    inputs = map(_to_dict, inputs)

    joined_dict = {}
    for i in inputs:
        for key, value in i.items():
            joined_dict[key] = value
    return joined_dict


def flatten_to_list_of_dicts(inputs):
    inputs = _flatten(inputs)
    inputs = map(_to_dict, inputs)

    joined_dict = collections.defaultdict(list)
    for i in inputs:
        for key, value in i.items():
            joined_dict[key].append(value)
    return dict(joined_dict)


def task_iterator(task, only_non_complete=False):
    # create cache of already seen tasks, so that when we recurse through the
    # DAG and multiple nodes have the same child, we don't return
    # the same child multiple times
    already_seen_tasks = set()

    # create another private function for recursion that has reference to `already_seen_tasks`
    def _unique_task_iterator(task, only_non_complete):
        if only_non_complete and task.complete():
            return

        yield task

        for dep in task.deps():
            if dep.task_id in already_seen_tasks:
                continue
            already_seen_tasks.add(dep.task_id)
            yield from _unique_task_iterator(dep, only_non_complete=only_non_complete)

    yield from _unique_task_iterator(task, only_non_complete)


def get_all_output_files_in_tree(root_module, key=None):
    if key:
        return get_all_output_files_in_tree(root_module)[key]

    all_output_files = collections.defaultdict(list)
    for task in task_iterator(root_module):
        output_dict = flatten_to_dict(task.output())
        if not output_dict:
            continue

        for target_key, target in output_dict.items():
            converted_dict = flatten_to_file_paths({target_key: target})
            file_key, file_name = converted_dict.popitem()

            all_output_files[file_key].append(
                dict(
                    exists=target.exists(),
                    parameters=get_serialized_parameters(task),
                    file_name=os.path.abspath(file_name),
                )
            )

    return all_output_files


def filter_from_params(output_files, **kwargs):
    kwargs_list = fill_kwargs_with_lists(**kwargs)

    if not kwargs_list:
        return output_files

    file_names = []

    for kwargs in product_dict(**kwargs_list):
        for output_dict in output_files:
            parameters = output_dict["parameters"]

            not_use = False
            for key, value in kwargs.items():
                if key in parameters and str(parameters[key]) != str(value):
                    not_use = True
                    break

            if not_use:
                continue

            file_names.append(output_dict)

    return {x["file_name"]: x for x in file_names}.values()


def get_task_from_file(file_name, task_name, **kwargs):
    spec = importlib.util.spec_from_file_location("module.name", os.path.basename(file_name))
    task_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(task_module)

    m = getattr(task_module, task_name)(**kwargs)

    return m


def get_serialized_parameters(task):
    """Get a string-typed ordered dict of key=value for the significant parameters"""
    serialized_parameters = collections.OrderedDict()

    for key, parameter in task.get_params():
        if not parameter.significant:
            continue

        value = getattr(task, key)
        if hasattr(parameter, "serialize_hashed"):
            value = parameter.serialize_hashed(value)
        else:
            value = parameter.serialize(value)

        serialized_parameters[key] = value

    return serialized_parameters


def create_output_file_name(task, base_filename, result_dir=None):
    serialized_parameters = get_serialized_parameters(task)

    if not result_dir:
        # Be sure to evaluate things relative to the current executed file, not to where we are now
        result_dir = map_folder(get_setting("result_dir", task=task, default=".", deprecated_keys=["result_path"]))

    for key, value in serialized_parameters.items():
        # Raise error if parameter value contains path separator "/" ("\" on Windows)
        # or is not interpretable as basename due to other reasons.
        if value != os.path.basename(value):
            raise ValueError(
                f"Parameter `{key}={value}` cannot be interpreted as directory name. "
                f"Make sure it does not contain the path separators ``{os.path.sep}``. "
                "Consider using a hashed parameter (e.g. ``b2luigi.Parameter(hashed=True)``)."
            )

    param_list = [f"{key}={value}" for key, value in serialized_parameters.items()]
    output_path = os.path.join(result_dir, *param_list)

    return os.path.join(output_path, base_filename)


def get_log_file_dir(task):
    if hasattr(task, "get_log_file_dir"):
        log_file_dir = task.get_log_file_dir()
        return log_file_dir

    # Be sure to evaluate things relative to the current executed file, not to where we are now
    base_log_file_dir = map_folder(get_setting("log_dir", task=task, default="logs", deprecated_keys=["log_folder"]))
    log_file_dir = create_output_file_name(task, task.get_task_family() + "/", result_dir=base_log_file_dir)

    return log_file_dir


def get_task_file_dir(task):
    if hasattr(task, "get_task_file_dir"):
        task_file_dir = task.get_task_file_dir()
        return task_file_dir

    task_file_dir = create_output_file_name(task, task.get_task_family() + "/")

    return task_file_dir


def get_filename():
    import __main__

    return __main__.__file__


def map_folder(input_folder):
    if os.path.isabs(input_folder):
        return input_folder

    try:
        filename = get_filename()
    except AttributeError as ex:
        raise type(ex)(
            "Could not determine the current script location. "
            "If you are running in an interactive shell (such as jupyter notebook) "
            "make sure to only provide absolute paths in your settings. "
            f"More Info:\n{ex}"
        ).with_traceback(sys.exc_info()[2])

    filepath = os.path.dirname(filename)

    return os.path.join(filepath, input_folder)


def _to_dict(d):
    if isinstance(d, dict):
        return d

    return {d: d}


def _flatten(struct):
    if isinstance(struct, dict) or isinstance(struct, str):
        return [struct]

    result = []
    try:
        iterator = iter(struct)
    except TypeError:
        return [struct]

    for f in iterator:
        result += _flatten(f)

    return result


def on_failure(self, exception):
    log_file_dir = os.path.abspath(get_log_file_dir(self))

    print(colorama.Fore.RED)
    print("Task", self.task_family, "failed!")
    print("Parameters")
    for key, value in get_filled_params(self).items():
        print("\t", key, "=", value)
    print("Please have a look into the log files in")
    print(log_file_dir)
    print(colorama.Style.RESET_ALL)


def add_on_failure_function(task):
    task.on_failure = types.MethodType(on_failure, task)


def create_cmd_from_task(task):
    filename = os.path.basename(get_filename())

    prefix = get_setting("executable_prefix", task=task, default=[], deprecated_keys=["cmd_prefix"])

    if isinstance(prefix, str):
        raise ValueError("Your specified executable_prefix needs to be a list of strings, e.g. [strace]")

    cmd = prefix

    executable = get_setting("executable", task=task, default=[sys.executable])

    if isinstance(executable, str):
        raise ValueError("Your specified executable needs to be a list of strings, e.g. [python3]")

    cmd += executable
    cmd += [filename, "--batch-runner", "--task-id", task.task_id]

    return cmd


def create_output_dirs(task):
    """Create all output dicts if needed. Normally only used internally."""
    output_list = flatten_to_dict(task.output())
    output_list = output_list.values()

    for output in output_list:
        output.makedirs()


def get_filled_params(task):
    """Helper function for getting the parameter list with each parameter set to its current value"""
    return {key: getattr(task, key) for key, _ in task.get_params()}



from collections import namedtuple
from glob import glob
import enum

from b2luigi.core.utils import product_dict, fill_kwargs_with_lists

import b2luigi

from parse import parse


class DataMode(enum.Enum):
    raw = "raw"
    mdst = "mdst"
    cdst = "cdst"
    skimmed_raw = "skimmed_raw"


class DataTask(b2luigi.ExternalTask):
    data_mode = b2luigi.EnumParameter(enum=DataMode)
    experiment_number = b2luigi.IntParameter()
    run_number = b2luigi.IntParameter()
    prefix = b2luigi.Parameter()
    file_name = b2luigi.Parameter()


class RawDataTask(DataTask):
    data_mode = DataMode.raw

    def output(self):
        yield {"raw_output.root": b2luigi.LocalTarget(_build_data_path(self))}


class DstDataTask(DataTask):
    release = b2luigi.Parameter()
    prod = b2luigi.IntParameter()
    database = b2luigi.IntParameter()

    def output(self):
        yield {"full_output.root": b2luigi.LocalTarget(_build_data_path(self))}


class SkimmedRawDataTask(DstDataTask):
    data_mode = DataMode.skimmed_raw

    def output(self):
        yield {"raw_output.root": b2luigi.LocalTarget(_build_data_path(self))}


class MdstDataTask(DstDataTask):
    data_mode = DataMode.mdst

    def output(self):
        yield {"output.root": b2luigi.LocalTarget(_build_data_path(self))}


class CdstDataTask(DstDataTask):
    data_mode = DataMode.cdst

    def output(self):
        yield {"output.root": b2luigi.LocalTarget(_build_data_path(self))}


requires_raw_data = b2luigi.requires(RawDataTask)
requires_skimmed_raw_data = b2luigi.requires(SkimmedRawDataTask)
requires_mdst_data = b2luigi.requires(MdstDataTask)
requires_cdst_data = b2luigi.requires(CdstDataTask)


def _get_dir_structure(data_mode):
    if data_mode == DataMode.mdst:
        return b2luigi.get_setting(
            "mdst_dir_structure",
            "/hsm/belle2/bdata/Data/release-{p.release}/DB{p.database:08d}/prod{p.prod:08d}/"
            + "e{p.experiment_number:04d}/4S/r{p.run_number:05d}/all/mdst/sub00/"
            + "mdst.{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root",
        )
    if data_mode == DataMode.cdst:
        return b2luigi.get_setting(
            "cdst_dir_structure",
            "/hsm/belle2/bdata/Data/release-{p.release}/DB{p.database:08d}/prod{p.prod:08d}/"
            + "e{p.experiment_number:04d}/4S/r{p.run_number:05d}/all/cdst/sub00/"
            + "cdst.{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root",
        )
    if data_mode == DataMode.skimmed_raw:
        return b2luigi.get_setting(
            "skimmed_raw_dir_structure",
            "/hsm/belle2/bdata/Data/release-{p.release}/DB{p.database:08d}/prod{p.prod:08d}/"
            + "e{p.experiment_number:04d}/4S/r{p.run_number:05d}/all/raw/sub00/"
            + "raw.{p.prefix}.{p.file_name}.{p.experiment_number:04d}.{p.run_number:05d}.root",
        )
    if data_mode == DataMode.raw:
        return b2luigi.get_setting(
            "raw_dir_structure",
            "/ghi/fs01/belle2/bdata/Data/Raw/e{p.experiment_number:04d}/r{p.run_number:05d}/sub00/"
            + "{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root",
        )


def _build_data_path(parameters):
    mode = parameters.data_mode

    dir_structure = _get_dir_structure(mode)
    return dir_structure.format(p=parameters)


def _parse_data_path(data_mode, data_path):
    extracted_kwargs = None

    dir_structure = _get_dir_structure(data_mode)
    extracted_kwargs = parse(dir_structure, data_path)

    if not extracted_kwargs:
        raise ValueError("The path does not fit the given format")
    extracted_kwargs = {key.replace("p.", ""): value for key, value in extracted_kwargs.named.items()}
    return extracted_kwargs


def _get_data_kwargs(data_mode, experiment_number, run_number, prefix=None, file_name=None, **other_kwargs):
    if file_name is None:
        file_name = "*"

    if prefix is None:
        prefix = "*"

    all_kwargs = fill_kwargs_with_lists(
        data_mode=data_mode,
        experiment_number=experiment_number,
        run_number=run_number,
        prefix=prefix,
        file_name=file_name,
        **other_kwargs,
    )
    for kwargs in product_dict(**all_kwargs):
        # The build_data_path wants an object, so lets convert the dict to a named tuple
        kwargs = namedtuple("GenericDict", kwargs.keys())(**kwargs)
        for data_file in glob(_build_data_path(kwargs)):
            yield _parse_data_path(data_mode, data_file)


def clone_on_mdst(
    self,
    task_class,
    experiment_number,
    run_number,
    release,
    prod,
    database,
    prefix=None,
    file_name=None,
    **additional_kwargs,
):
    # TODO: make database not needed
    for kwargs in _get_data_kwargs(
        data_mode=DataMode.mdst,
        experiment_number=experiment_number,
        run_number=run_number,
        release=release,
        prod=prod,
        database=database,
        prefix=prefix,
        file_name=file_name,
    ):
        yield self.clone(task_class, **kwargs, **additional_kwargs)


def clone_on_cdst(
    self,
    task_class,
    experiment_number,
    run_number,
    release,
    prod,
    database,
    prefix=None,
    file_name=None,
    **additional_kwargs,
):
    # TODO: make database not needed
    for kwargs in _get_data_kwargs(
        data_mode=DataMode.cdst,
        experiment_number=experiment_number,
        run_number=run_number,
        release=release,
        prod=prod,
        database=database,
        prefix=prefix,
        file_name=file_name,
    ):
        yield self.clone(task_class, **kwargs, **additional_kwargs)


def clone_on_skimmed_raw(
    self,
    task_class,
    experiment_number,
    run_number,
    release,
    prod,
    database,
    prefix=None,
    file_name=None,
    **additional_kwargs,
):
    # TODO: make database not needed
    for kwargs in _get_data_kwargs(
        data_mode=DataMode.skimmed_raw,
        experiment_number=experiment_number,
        run_number=run_number,
        release=release,
        prod=prod,
        database=database,
        prefix=prefix,
        file_name=file_name,
    ):
        yield self.clone(task_class, **kwargs, **additional_kwargs)


def clone_on_raw(self, task_class, experiment_number, run_number, prefix=None, file_name=None, **additional_kwargs):
    for kwargs in _get_data_kwargs(
        data_mode=DataMode.raw,
        experiment_number=experiment_number,
        run_number=run_number,
        prefix=prefix,
        file_name=file_name,
    ):
        yield self.clone(task_class, **kwargs, **additional_kwargs)



import argparse


def get_cli_arguments(ignore_additional_command_line_args=False):
    """

    Args:
        ignore_additional_command_line_args (bool, optional, default False): Ignore additional
            command line arguments. This is useful if you want to use this function in a file
            that also does some command line parsing.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--show-output",
        help="Instead of running the tasks, show which output files will/are created.",
        action="store_true",
    )
    parser.add_argument(
        "--test",
        help="Run the task list in test mode by printing the log directly to the screen instead"
        " of storing it in a file.",
        action="store_true",
    )
    parser.add_argument(
        "--batch", help="Instead of running locally, try to submit the tasks to the batch system.", action="store_true"
    )
    parser.add_argument(
        "--batch-runner", help="Expert option to mark this worker as a batch runner.", action="store_true"
    )
    parser.add_argument(
        "--dry-run",
        help="Do not run any task but set the return value to 0, if the tasks are complete.",
        action="store_true",
    )
    parser.add_argument(
        "--scheduler-host", help="If given, use this host as a central scheduler instead of a local one.", default=""
    )
    parser.add_argument(
        "--scheduler-port",
        help="If given, use the port on this host as a central scheduler instead of a local one.",
        type=int,
        default=0,
    )

    parser.add_argument("--task-id", help="EXPERT.", default="")

    if not ignore_additional_command_line_args:
        args = parser.parse_args()
    else:
        args, _ = parser.parse_known_args()

    if (args.test or args.dry_run) and (args.scheduler_host or args.scheduler_port):
        raise AttributeError("Can not test while using a central scheduler!")
    if args.batch_runner and not args.task_id:
        raise AttributeError("A batch runner should always have a fully qualified task id.")
    if args.show_output and (args.scheduler_host or args.scheduler_port or args.batch or args.test):
        print("Ignoring all other parameters, as you have given the --show-output parameter.")

    return args



import json
import os
import tempfile
import unittest
from collections import Counter
from typing import List
from unittest.mock import MagicMock, Mock, patch

import b2luigi
from b2luigi.batch.processes.gbasf2 import Gbasf2Process, JobStatus, get_unique_project_name

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask


class MyGbasf2Task(MyTask):
    gbasf2_project_name_prefix = "my_gb2_task_"


class TestGbasf2RescheduleJobs(B2LuigiTestCase):
    job_statuses_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "_gbasf2_project_statuses")
    joblist_tmpfile_name = "jobs_to_be_rescheduled.txt"

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = MagicMock()
        self.gb2_mock_process.task = MyGbasf2Task("some_parameter")
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.gbasf2_project_name = get_unique_project_name(self.gb2_mock_process.task)
        self.gb2_mock_process.n_retries_by_job = Counter()
        self.gb2_mock_process.max_retries = 1
        b2luigi.set_setting("gbasf2_print_status_updates", False)

    def _get_job_status_dict(self, job_status_fname):
        with open(os.path.join(self.job_statuses_dir, job_status_fname)) as job_status_json_file:
            return json.load(job_status_json_file)

    def _reschedule_jobs(self, job_ids):
        pass

    def assert_rescheduled_jobs(self, job_status_fname, expected_jobs_to_be_rescheduled):
        with patch(
            "b2luigi.batch.processes.gbasf2.get_gbasf2_project_job_status_dict",
            MagicMock(return_value=self._get_job_status_dict(job_status_fname)),
        ):
            with patch("b2luigi.batch.processes.gbasf2.Gbasf2Process._reschedule_jobs", new=self._reschedule_jobs):
                Gbasf2Process._reschedule_failed_jobs(self.gb2_mock_process)

                self.assertCountEqual(
                    list(self.gb2_mock_process.n_retries_by_job.keys()), expected_jobs_to_be_rescheduled
                )
                for jobid in self.gb2_mock_process.n_retries_by_job.keys():
                    self.assertEqual(self.gb2_mock_process.n_retries_by_job[jobid], 1)

    def test_reschedule_jobs_all_done(self):
        "Test gbasf2 project rescheduling with dict where all jobs are done"
        self.assert_rescheduled_jobs("done_testjbucket1357828d80b3.json", [])

    def test_reschedule_jobs_one_failed(self):
        "Test gbasf2 project rescheduling with dict where one job in project failed"
        self.assert_rescheduled_jobs("failed_7663_r03743_10_prod00013766_11x1.json", ["188623842", "188625261"])

    def test_reschedule_jobs_running(self):
        "Test gbasf2 project rescheduling with dict where several jobs are either running or waiting"
        self.assert_rescheduled_jobs("running_TrainingFEI_17-04-2021a10a957289.json", [])

    def test_reschedule_jobs_major_status_done_but_minor_status_not(self):
        "Test gbasf2 project rescheduling with dict where major job status is all done but application status is not ``DONE``"
        self.assert_rescheduled_jobs("all_done_but_application_error.json", ["187522107"])


class TestGbasf2GetJobStatus(B2LuigiTestCase):
    job_statuses_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "_gbasf2_project_statuses")

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = Mock()
        self.gb2_mock_process.task = MyGbasf2Task("some_parameter")
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.gbasf2_project_name = get_unique_project_name(self.gb2_mock_process.task)
        self.gb2_mock_process.max_retries = 0
        b2luigi.set_setting("gbasf2_print_status_updates", False)

    def _get_job_status_dict(self, job_status_fname):
        with open(os.path.join(self.job_statuses_dir, job_status_fname)) as job_status_json_file:
            return json.load(job_status_json_file)

    def assert_job_status(self, job_status_fname, expected_job_status):
        with patch(
            "b2luigi.batch.processes.gbasf2.get_gbasf2_project_job_status_dict",
            MagicMock(return_value=self._get_job_status_dict(job_status_fname)),
        ):
            job_status = Gbasf2Process.get_job_status(self.gb2_mock_process)
            self.assertEqual(job_status, expected_job_status)

    def test_get_job_status_all_done(self):
        "Test gbasf2 project status dict where all jobs are done"
        self.assert_job_status("done_testjbucket1357828d80b3.json", JobStatus.successful)

    def test_get_job_status_one_failed(self):
        "Test gbasf2 project status dict where one job in project failed"
        self.assert_job_status("failed_7663_r03743_10_prod00013766_11x1.json", JobStatus.aborted)

    def test_get_job_status_running(self):
        "Test gbasf2 project status dict where several jobs are either running or waiting"
        self.assert_job_status("running_TrainingFEI_17-04-2021a10a957289.json", JobStatus.running)

    def test_get_job_status_major_status_done_but_minor_status_not(self):
        "Test gbasf2 project status dict where major job status is all done but application status is not ``DONE``"
        self.assert_job_status("all_done_but_application_error.json", JobStatus.aborted)


class TestBuildGbasf2SubmitCommand(B2LuigiTestCase):
    dummy_lfn = "/grid/path/to/dataset"

    def setUp(self):
        super().setUp()
        self.dummy_lfn_file = tempfile.NamedTemporaryFile()

    def _get_task_file_dir(self, task):
        return self.test_dir

    def _build_gbasf2_submit_command(self, task) -> List[str]:
        gb2_mock_process = Mock()
        os.makedirs(self._get_task_file_dir(task), exist_ok=True)
        self.pickle_file_path = os.path.join(self._get_task_file_dir(task), "serialized_basf2_path.pkl")
        self.wrapper_file_path = os.path.join(self._get_task_file_dir(task), "steering_file_wrapper.py")
        gb2_mock_process.pickle_file_path = self.test_dir
        gb2_mock_process.wrapper_file_path = self.test_dir
        gb2_mock_process.gbasf2_project_name = get_unique_project_name(task)
        gb2_mock_process.task = task
        return Gbasf2Process._build_gbasf2_submit_command(gb2_mock_process)

    def test_input_dataset_in_command(self):
        "Test that if the input dataset is set, it occurs once in the command string"

        class _Task(MyGbasf2Task):
            gbasf2_input_dataset = b2luigi.Parameter(default=self.dummy_lfn)

        gb2_cmd = self._build_gbasf2_submit_command(_Task("some_parameter"))
        gb2_cmd_str = " ".join(gb2_cmd)
        self.assertEqual(gb2_cmd.count(self.dummy_lfn), 1)
        self.assertEqual(gb2_cmd.count("-i"), 1)
        self.assertEqual(gb2_cmd_str.count(f" -i {self.dummy_lfn}"), 1)
        self.assertNotIn("--input_dslist", gb2_cmd_str)

    def test_input_dslist_in_command(self):
        "Test that if the input dataset list is set, it occurs once in the command string"

        class _Task(MyGbasf2Task):
            gbasf2_input_dslist = self.dummy_lfn_file.name

        gb2_cmd = self._build_gbasf2_submit_command(_Task("some_parameter"))
        gb2_cmd_str = " ".join(gb2_cmd)
        self.assertEqual(gb2_cmd.count(self.dummy_lfn_file.name), 1)
        self.assertEqual(gb2_cmd.count("--input_dslist"), 1)
        self.assertEqual(gb2_cmd_str.count(f" --input_dslist {self.dummy_lfn_file.name}"), 1)
        self.assertNotIn("-i", gb2_cmd)

    def test_input_dslist_set_but_does_not_exist(self):
        """Test that if the input dataset list is set but the file does not exist, an ``FileNotFoundError`` is raised"""

        class _Task(MyGbasf2Task):
            gbasf2_input_dslist = self.dummy_lfn_file.name

        self.dummy_lfn_file.close()
        self.assertRaises(FileNotFoundError, lambda: self._build_gbasf2_submit_command(_Task("some_parameter")))

    def test_not_dataset_raises_error(self):
        """Test that we get an error when no gbasf2 input dataset is provided"""
        self.assertRaises(RuntimeError, lambda: self._build_gbasf2_submit_command(MyGbasf2Task("some_parameter")))

    def test_both_input_dataset_and_dslist_raises_error(self):
        """Test that en error is raise when both gbasf2_input_dataset and gbasf2_inputdslist settings are given"""

        class _Task(MyGbasf2Task):
            gbasf2_input_dataset = b2luigi.Parameter(default=self.dummy_lfn)
            gbasf2_input_dslist = self.dummy_lfn_file.name

        self.assertRaises(RuntimeError, lambda: self._build_gbasf2_submit_command(MyGbasf2Task("some_parameter")))


class TestGetGbasf2DatasetQuery(unittest.TestCase):
    "Run tests for ``Gbasf2Process._get_gbasf2_dataset_query``."

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = Mock()
        self.gb2_mock_process.gbasf2_project_name = "projectname"
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.task = MyTask("some_parameter")

    def test_filename_single_root_extension(self):
        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="output.root")
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.root")

    def test_filename_multi_extension(self):
        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(
            self.gb2_mock_process, output_file_name="output.mdst.root"
        )
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.mdst.root")

    def test_non_root_extension_raises_err(self):
        "Raise an error if gbasf2 output does not end with .root"
        with self.assertRaises(ValueError):
            Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="output.mdst")

    def test_non_basename_raises_err(self):
        """
        Currently we only support simple basenames as gbasf2 job outputs.
        LPN is guessed via dirac username for user projects.
        """
        with self.assertRaises(ValueError):
            Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="/path/to/output.root")

    def test_filename_single_extension(self):
        gb2_mock_process = Mock()
        gb2_mock_process.gbasf2_project_name = "projectname"
        gb2_mock_process.dirac_user = "username"
        gb2_mock_process.task = MyTask("some_parameter")

        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(self=gb2_mock_process, output_file_name="output.root")
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.root")



#!/usr/bin/env python3
"""
This file is not a standalone example, but meant to provide a simple example
analysis path be imported from gbasf2_example.py
"""

import basf2
import modularAnalysis as mA
import vertex as vx
from stdCharged import stdK, stdPi
from variables import variables as vm


def create_analysis_path(
    b_ntuple_filename="B_ntuple.root",
    d_ntuple_filename="D_ntuple.root",
    mbc_range=(5.2, 5.3),
):
    """
    Example of a minimal reconstruction with a cut as a changeable function
    parameter, adapted from code in the ``B2T_Basics_3_FirstAnalysis.ipynb``
    notebook from b2 starter kit.
    """
    path = basf2.Path()
    # this local inputMdstList will only be used when this steerig file is run locally, gbasf2 overrides it
    local_input_files = [
        "/group/belle2/dataprod/MC/MC13a/prod00009434/s00/e1003/4S/r00000/mixed/mdst/sub00/mdst_000001_prod00009434_task10020000001.root"
    ]
    mA.inputMdstList(
        environmentType="default",
        filelist=local_input_files,
        path=path,
    )
    stdK("higheff", path=path)
    stdPi("higheff", path=path)
    mA.reconstructDecay("D0:Kpi -> K-:higheff pi+:higheff", "1.7 < M < 1.9", path=path)
    # use try except to have this code work for both the old and new function names for the tree fit
    mA.matchMCTruth("D0:Kpi", path=path)
    mA.reconstructDecay("B- -> D0:Kpi pi-:higheff", f"{mbc_range[0]} < Mbc < {mbc_range[1]}", path=path)
    try:
        vx.treeFit("B+", 0.1, path=path)
    except AttributeError:
        vx.vertexTree("B+", 0.1, path=path)
    mA.setAnalysisConfigParams({"mcMatchingVersion": "BelleII"}, path)
    mA.matchMCTruth("B-", path=path)
    vm.addAlias("p_cms", "useCMSFrame(p)")  # include aliases to test if they work
    vm.addAlias("E_cms", "useCMSFrame(E)")
    mA.variablesToNtuple(
        "D0:Kpi",
        ["M", "p", "E", "E_cms", "p_cms", "daughter(0, kaonID)", "daughter(1, pionID)", "isSignal", "mcErrors"],
        filename=d_ntuple_filename,
        treename="D",
        path=path,
    )
    mA.variablesToNtuple(
        "B-", ["Mbc", "deltaE", "isSignal", "mcErrors", "M"], filename=b_ntuple_filename, treename="B", path=path
    )
    return path


if __name__ == "__main__":
    # Use this to execute the analysis path if this file is called as a main
    # file. It can be used to test the analysis path independently of the gbasf2
    # luigi task. But if this module is only imported, the following is not executed.
    path = create_analysis_path()
    basf2.print_path(path)
    basf2.process(path)



import os
import subprocess

from ..helpers import B2LuigiTestCase


class BatchProcessTestCase(B2LuigiTestCase):
    def test_simple_task(self):
        self.call_file("batch/batch_task_1.py")
        self.assertTrue(os.path.exists("some_parameter=bla_blub/test.txt"))
        self.assertTrue(os.path.exists("some_parameter=bla_blub/combined.txt"))

    def test_failing_task(self):
        out = self.call_file("batch/batch_task_2.py", stderr=subprocess.STDOUT)
        self.assertTrue(os.path.exists("some_parameter=bla_blub/test.txt"))
        self.assertFalse(os.path.exists("some_parameter=bla_blub/combined.txt"))

        self.assertIn(b"Task MyAdditionalTask failed!", out.splitlines())
        self.assertIn(b"Please have a look into the log files in", out.splitlines())
        self.assertIn(b"This progress looks :( because there were failed tasks", out.splitlines())



import b2luigi


class MyTask(b2luigi.Task):
    some_parameter = b2luigi.Parameter()

    def output(self):
        yield self.add_to_output("test.txt")

    @b2luigi.on_temporary_files
    def run(self):
        with open(self.get_output_file_name("test.txt"), "w") as f:
            f.write("Test")


@b2luigi.requires(MyTask)
class MyAdditionalTask(b2luigi.Task):
    def output(self):
        yield self.add_to_output("combined.txt")

    @b2luigi.dispatch
    @b2luigi.on_temporary_files
    def run(self):
        with open(self.get_output_file_name("combined.txt"), "w") as f:
            with open(self.get_input_file_names("test.txt")[0], "r") as from_f:
                f.write(from_f.read())


if __name__ == "__main__":
    b2luigi.set_setting("batch_system", "test")
    b2luigi.process(MyAdditionalTask(some_parameter="bla_blub"), batch=True)



import os

from ..helpers import B2LuigiTestCase

import b2luigi


class TemporaryWrapperTestCase(B2LuigiTestCase):
    def test_failed_temporary_files(self):
        self.call_file("core/temporary_task_1.py")

        self.assertFalse(os.path.exists("test.txt"))

    def test_good_temporary_files(self):
        self.call_file("core/temporary_task_2.py")

        self.assertTrue(os.path.exists("test.txt"))

        with open("test.txt", "r") as f:
            self.assertEqual(f.read(), "Test")

    def test_reset_output(self):
        class MyTask(b2luigi.Task):
            def output(self):
                yield self.add_to_output("test.txt")

            @b2luigi.on_temporary_files
            def run(self):
                with open(self.get_output_file_name("test.txt"), "w") as f:
                    f.write("Test")

        task = MyTask()

        self.assertNotIn("tmp", task.get_output_file_name("test.txt"))



import b2luigi as luigi
import random


class TaskA(luigi.Task):
    central_value = luigi.FloatParameter()
    index = luigi.IntParameter()

    def run(self):
        with open(self.get_output_file_name("random_numbers.txt"), "w") as f:
            for _ in range(1000):
                f.write(f"{random.gauss(self.central_value, 0.0)}\n")

    def output(self):
        yield self.add_to_output("random_numbers.txt")


class TaskB(luigi.Task):
    def requires(self):
        self.required_tasks = {
            "a": (TaskA(5.0, i) for i in range(100)),
            "b": (TaskA(1.0, i) for i in range(100)),
        }
        return self.required_tasks

    @staticmethod
    def average_file(input_file):
        # Build the mean
        summed_numbers = 0
        counter = 0
        with open(input_file, "r") as f:
            for line in f.readlines():
                summed_numbers += float(line)
                counter += 1

        return summed_numbers / counter

    def average_inputs(self, input_file_list):
        summed_averages = 0
        counter = 0
        for input_file in input_file_list:
            summed_averages += self.average_file(input_file)
            counter += 1

        return summed_averages / counter

    def run(self):
        print(self.get_input_file_names_from_dict("a"))

        average_a = self.average_inputs(self.get_input_file_names_from_dict("a", "random_numbers.txt"))
        average_b = self.average_inputs(self.get_input_file_names_from_dict("b")["random_numbers.txt"])

        with open(self.get_output_file_name("combined_average.txt"), "w") as out_file:
            out_file.write(f"{(average_a + average_b)/2}\n")

    def output(self):
        yield self.add_to_output("combined_average.txt")


if __name__ == "__main__":
    luigi.set_setting("result_path", "results")
    luigi.process(TaskB(), workers=4)



import json
import os
import re
import subprocess
import enum

from retry import retry

from b2luigi.core.settings import get_setting
from b2luigi.batch.processes import BatchProcess, JobStatus
from b2luigi.batch.cache import BatchJobStatusCache
from b2luigi.core.utils import get_log_file_dir, get_task_file_dir
from b2luigi.core.executable import create_executable_wrapper


class HTCondorJobStatusCache(BatchJobStatusCache):
    @retry(subprocess.CalledProcessError, tries=3, delay=2, backoff=3)  # retry after 2,6,18 seconds
    def _ask_for_job_status(self, job_id: int = None):
        """
        With HTCondor, you can check the progress of your jobs using the `condor_q` command.
        If no `JobId` is given as argument, this command shows you the status of all queued jobs
        (usually only your own by default).

        Normally the HTCondor `JobID` is stated as `ClusterId.ProcId`. Since only on job is queued per
        cluster, we can identify jobs by their `ClusterId` (The `ProcId` will be 0 for all submitted jobs).
        With the `-json` option, the `condor_q` output is returned in the JSON format. By specifying some
        attributes, not the entire job ClassAd is returned, but only the necessary information to match a
        job to its `JobStatus`. The output is given as `string` and cannot be directly parsed into a json
        dictionary. It has the following form:
            [
                {...}
                ,
                {...}
                ,
                {...}
            ]
        The {...} are the different dictionaries including the specified attributes.
        Sometimes it might happen that a job is completed in between the status checks. Then its final status
        can be found in the `condor_history` file (works mostly in the same way as `condor_q`).
        Both commands are used in order to find out the `JobStatus`.
        """
        # https://htcondor.readthedocs.io/en/latest/man-pages/condor_q.html
        q_cmd = ["condor_q", "-json", "-attributes", "ClusterId,JobStatus,ExitStatus"]

        if job_id:
            output = subprocess.check_output(q_cmd + [str(job_id)])
        else:
            output = subprocess.check_output(q_cmd)

        seen_ids = self._fill_from_output(output)

        # If the specified job can not be found in the condor_q output, we need to request its history
        if job_id and job_id not in seen_ids:
            # https://htcondor.readthedocs.io/en/latest/man-pages/condor_history.html
            history_cmd = [
                "condor_history",
                "-json",
                "-attributes",
                "ClusterId,JobStatus,ExitCode",
                "-match",
                "1",
                str(job_id),
            ]
            output = subprocess.check_output(history_cmd)

            self._fill_from_output(output)

    def _fill_from_output(self, output):
        output = output.decode()

        seen_ids = set()

        if not output:
            return seen_ids

        for status_dict in json.loads(output):
            if status_dict["JobStatus"] == HTCondorJobStatus.completed and status_dict["ExitCode"]:
                self[status_dict["ClusterId"]] = HTCondorJobStatus.failed
            else:
                self[status_dict["ClusterId"]] = status_dict["JobStatus"]

            seen_ids.add(status_dict["ClusterId"])

        return seen_ids


class HTCondorJobStatus(enum.IntEnum):
    """
    See https://htcondor.readthedocs.io/en/latest/classad-attributes/job-classad-attributes.html
    """

    idle = 1
    running = 2
    removed = 3
    completed = 4
    held = 5
    transferring_output = 6
    suspended = 7
    failed = 999


_batch_job_status_cache = HTCondorJobStatusCache()


class HTCondorProcess(BatchProcess):
    """
    Reference implementation of the batch process for a HTCondor batch system.

    Additional to the basic batch setup (see :ref:`batch-label`), additional
    HTCondor-specific things are:

    * Please note that most of the HTCondor batch farms do not have the same
      environment setup on submission and worker machines, so you probably want to give an
      ``env_script``, an ``env`` :meth:`setting <b2luigi.set_setting>` and/or a different ``executable``.

    * HTCondor supports copying files from submission to workers. This means if the
      folder of your script(s)/python project/etc. is not accessible on the worker, you can
      copy it from the submission machine by adding it to the setting ``transfer_files``.
      This list can host both folders and files.
      Please note that due to HTCondors file transfer mechanism, all specified folders
      and files will be copied into the worker node flattened, so if you specify
      `a/b/c.txt` you will end up with a file `c.txt`.
      If you use the ``transfer_files`` mechanism, you need to set the ``working_dir`` setting to "."
      as the files will end up in the current worker scratch folder.
      All specified files/folders should be absolute paths.

      .. hint::
        Please do not specify any parts or the full results folder. This will lead to unexpected
        behavior. We are working on a solution to also copy results, but until this the
        results folder is still expected to be shared.

      If you copy your python project using this setting to the worker machine, do not
      forget to actually set it up in your setup script.
      Additionally, you might want to copy your ``settings.json`` as well.

    * Via the ``htcondor_settings`` setting you can provide a dict as
      a for additional options, such as requested memory etc. Its value has to be a dictionary
      containing HTCondor settings as key/value pairs. These options will be written into the job
      submission file. For an overview of possible settings refer to the `HTCondor documentation
      <https://htcondor.readthedocs.io/en/latest/users-manual/submitting-a-job.html#>`_.

    * Same as for the :ref:`lsf`, the ``job_name`` setting allows giving a meaningful name to a
      group of jobs. If you want to be htcondor-specific, you can provide the ``JobBatchName`` as an
      entry in the ``htcondor_settings`` dict, which will override the global ``job_name`` setting.
      This is useful for manually checking the status of specific jobs with

      .. code-block:: bash

        condor_q -batch <job name>

    Example:

        .. literalinclude:: ../../examples/htcondor/htcondor_example.py
           :linenos:
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_job_id = None

    def get_job_status(self):
        if not self._batch_job_id:
            return JobStatus.aborted

        try:
            job_status = _batch_job_status_cache[self._batch_job_id]
        except KeyError:
            return JobStatus.aborted

        if job_status in [HTCondorJobStatus.completed]:
            return JobStatus.successful
        if job_status in [
            HTCondorJobStatus.idle,
            HTCondorJobStatus.running,
            HTCondorJobStatus.transferring_output,
            HTCondorJobStatus.suspended,
        ]:
            return JobStatus.running
        if job_status in [HTCondorJobStatus.removed, HTCondorJobStatus.held, HTCondorJobStatus.failed]:
            return JobStatus.aborted
        raise ValueError(f"Unknown HTCondor Job status: {job_status}")

    def start_job(self):
        submit_file = self._create_htcondor_submit_file()

        # HTCondor submit needs to be called in the folder of the submit file
        submit_file_dir, submit_file = os.path.split(submit_file)
        output = subprocess.check_output(["condor_submit", submit_file], cwd=submit_file_dir)

        output = output.decode()
        match = re.search(r"[0-9]+\.", output)
        if not match:
            raise RuntimeError("Batch submission failed with output " + output)

        self._batch_job_id = int(match.group(0)[:-1])

    def terminate_job(self):
        if not self._batch_job_id:
            return

        subprocess.run(["condor_rm", str(self._batch_job_id)], stdout=subprocess.DEVNULL)

    def _create_htcondor_submit_file(self):
        submit_file_content = []

        # Specify where to write the log to
        log_file_dir = get_log_file_dir(self.task)
        os.makedirs(log_file_dir, exist_ok=True)

        stdout_log_file = os.path.abspath(os.path.join(log_file_dir, "stdout"))
        submit_file_content.append(f"output = {stdout_log_file}")

        stderr_log_file = os.path.abspath(os.path.join(log_file_dir, "stderr"))
        submit_file_content.append(f"error = {stderr_log_file}")

        job_log_file = os.path.abspath(os.path.join(log_file_dir, "job.log"))
        submit_file_content.append(f"log = {job_log_file}")

        # Specify the executable
        executable_file = create_executable_wrapper(self.task)
        submit_file_content.append(f"executable = {os.path.basename(executable_file)}")

        # Specify additional settings
        general_settings = get_setting("htcondor_settings", dict())
        try:
            general_settings.update(self.task.htcondor_settings)
        except AttributeError:
            pass

        transfer_files = get_setting("transfer_files", task=self.task, default=[])
        if transfer_files:
            working_dir = get_setting("working_dir", task=self.task, default="")
            if not working_dir or working_dir != ".":
                raise ValueError("If using transfer_files, the working_dir must be explicitely set to '.'")

            general_settings.setdefault("should_transfer_files", "YES")
            general_settings.setdefault("when_to_transfer_output", "ON_EXIT")

            transfer_files = set(transfer_files)

            for transfer_file in transfer_files:
                if os.path.abspath(transfer_file) != transfer_file:
                    raise ValueError(
                        "You should only give absolute file names in transfer_files!"
                        + f"{os.path.abspath(transfer_file)} != {transfer_file}"
                    )

            env_setup_script = get_setting("env_script", task=self.task, default="")
            if env_setup_script:
                # TODO: make sure to call it relatively
                transfer_files.add(os.path.abspath(env_setup_script))

            general_settings.setdefault("transfer_input_files", ",".join(transfer_files))

        job_name = get_setting("job_name", task=self.task, default=False)
        if job_name is not False:
            general_settings.setdefault("JobBatchName", job_name)

        for key, item in general_settings.items():
            submit_file_content.append(f"{key} = {item}")

        # Finally also start the process
        submit_file_content.append("queue 1")

        # Now we can write the submit file
        output_path = get_task_file_dir(self.task)
        submit_file_path = os.path.join(output_path, "job.submit")

        os.makedirs(output_path, exist_ok=True)

        with open(submit_file_path, "w") as submit_file:
            submit_file.write("\n".join(submit_file_content))

        return submit_file_path



import pickle

from basf2 import conditions as b2conditions, logging
from basf2.pickle_path import serialize_path
from variables import variables as vm


def get_alias_dict_from_variable_manager():
    """
    Extracts a dictionary with the alias names as keys and their values from the
    internal state of the variable manager and returns it.
    """
    alias_dictionary = {str(alias_name): vm.getVariable(str(alias_name)).name for alias_name in vm.getAliasNames()}
    return alias_dictionary


def write_path_and_state_to_file(basf2_path, file_path):
    """
    Serialize basf2 path, aliases and log_level to file.

    Variant of ``basf2.pickle_path.write_path_to_file``, only with additional
    serialization of the basf2 variable aliases and global tags.

    The aliases are extracted from the current state of the variable manager singleton
    and thus have to be added in the python/basf2 process before calling this function.
    Likewise for the global tags.

    :param path: Basf2 path object to serialize
    :param file_path: File path to write the serialized pickle object to.
    """
    with open(file_path, "bw") as pickle_file:
        serialized = serialize_path(basf2_path)
        serialized["aliases"] = get_alias_dict_from_variable_manager()
        serialized["globaltags"] = b2conditions.globaltags
        serialized["log_level"] = logging.log_level
        # serialized["conditions"] = b2conditions
        pickle.dump(serialized, pickle_file)



"""
Test helper functions for :py:class:`Gbasf2Process`.

Utilities that require a running gbasf2 or Dirac environment will not net
tested in this test case, only the functions that can be run independently.
"""

import json
import os
import subprocess
import unittest
from unittest import mock

import b2luigi
from b2luigi.batch.processes.htcondor import HTCondorJobStatusCache, HTCondorProcess

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask


class TestHTCondorCreateSubmitFile(B2LuigiTestCase):
    def _get_htcondor_submit_file_string(self, task):
        # the _create_htcondor_submit_file is a method of the ``HTCondorProcess`` class, but it only uses its
        # class to obtain ``self.task``, to it's sufficient to provide a mock class for ``self``, `
        htcondor_mock_process = mock.Mock()
        task.get_task_file_dir = lambda: self.test_dir
        task.get_log_file_dir = lambda: self.test_dir
        htcondor_mock_process.task = task
        #  create submit file
        HTCondorProcess._create_htcondor_submit_file(htcondor_mock_process)
        # read submit file and return string
        submit_file_path = os.path.join(self.test_dir, "job.submit")
        with open(submit_file_path, "r") as submit_file:
            return submit_file.read()

    def test_minimal_submit_file(self):
        """
        Minimal  submit file should have expected shape:

            output = ...
            error = ..
            log = ..
            executable = executable_wrapper.sh
            queue 1
        """
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        self.assertIn("output = ", submit_file_lines[0])
        self.assertIn("error = ", submit_file_lines[1])
        self.assertIn("log = ", submit_file_lines[2])
        self.assertEqual("executable = executable_wrapper.sh", submit_file_lines[3])
        self.assertEqual("queue 1", submit_file_lines[4])

    def test_not_setting_job_name(self):
        submit_file_string = self._get_htcondor_submit_file_string(MyTask("some_parameter"))
        self.assertNotIn("JobBatchName", submit_file_string)

    def test_set_job_name_via_task_attribute(self):
        task = MyTask("some_parameter")
        task.job_name = "some_job_name"
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

        b2luigi.set_setting("job_name", "some_job_name")
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        b2luigi.clear_setting("job_name")
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

    def test_set_job_name_is_overriden_by_htcondor_settings(self):
        """
        ``job_name`` is a global setting, but if the ``JobBatchName`` is set explicitly via the settings, we
        want that to override the global setting
        """
        task = MyTask("some_parameter")
        task.job_name = "job_name_global"
        htcondor_settings = {"JobBatchName": "job_name_htcondor"}
        b2luigi.set_setting("htcondor_settings", htcondor_settings)
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        b2luigi.clear_setting("htcondor_settings")
        self.assertNotIn("JobBatchName = job_name_global", submit_file_lines)
        self.assertIn("JobBatchName = job_name_htcondor", submit_file_lines)


class TestHTCondorJobStatusCache(unittest.TestCase):
    def setUp(self):
        mock_status_dicts = [
            {
                "JobStatus": 4,  # completed
                "ExitCode": 0,  # success
                "ClusterId": 42,
            },
            {
                "JobStatus": 2,  # running
                "ClusterId": 43,
            },
            {
                "JobStatus": 999,  # failed
                "ClusterId": 44,
            },
        ]
        self.mock_status_json = json.dumps(mock_status_dicts).encode()

        self.htcondor_job_status_cache = HTCondorJobStatusCache()

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_does_2_retries(self, mock_check_output):
        """Test the ``_ask_for_job_status`` recovers after two condor_q failures."""

        # make check_output fail 2 times before  return status dict
        n_fail = 2
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_json
        ]
        self.htcondor_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 3)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_no_retries_on_success(self, mock_check_output):
        """Test the ``_ask_for_job_status`` is only called once (no retries) when everything works."""

        # make check_output fail 2 times before  return status dict
        mock_check_output.return_value = self.mock_status_json
        self.htcondor_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 1)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_fails_after_4_condor_q_failures(self, mock_check_output):
        """Test the ``_ask_for_job_status`` does not do more than 3 retries"""

        # make check_output fail 2 times before  return status dict
        n_fail = 4
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_json
        ]
        with self.assertRaises(subprocess.CalledProcessError):
            self.htcondor_job_status_cache._ask_for_job_status()



from contextlib import ExitStack
from functools import wraps


class TemporaryFileContextManager(ExitStack):
    def __init__(self, task):
        super().__init__()

        self._task = task
        self._task_output_function = task.get_output_file_name

        self._open_files = {}

    def __enter__(self):
        def get_output_file_name(key):
            if key not in self._open_files:
                target = self._task._get_output_target(key)
                temporary_path = target.temporary_path()
                self._open_files[key] = self.enter_context(temporary_path)

            return self._open_files[key]

        self._task.get_output_file_name = get_output_file_name

    def __exit__(self, *exc_details):
        super().__exit__(*exc_details)

        self._task.get_output_file_name = self._task_output_function


def on_temporary_files(run_function):
    """
    Wrapper for decorating a task's run function to use temporary files as outputs.

    A common problem when using long running tasks in ``luigi`` is the so called thanksgiving bug
    (see https://www.arashrouhani.com/luigi-budapest-bi-oct-2015/#/21).
    It occurs, when you define an output of a task and in its run function,
    you create this output before filling it with content
    (maybe even only after a long lasting calculation).
    It may happen, that during the creation of the output and the finish of the calculation
    some other tasks checks if the output is already there, finds it and assumes,
    that the task is already finished (although there is probably only non-sense in the file
    so far).

    A solution is already given by luigi itself, when using the ``temporary_path()`` function
    of the file system targets, which is really nice!
    Unfortunately, this means you have to open all your output files with a context manager
    and this is very hard to do if you have external tasks also (because they will
    probably use the output file directly instead of the temporary file version of if).

    This wrapper simplifies the usage of the temporary files:

    .. code-block:: python

      import b2luigi

      class MyTask(b2luigi.Task):
          def output(self):
              yield self.add_to_output("test.txt")

          @b2luigi.on_temporary_files
          def run(self):
              with open(self.get_output_file_name("test.txt"), "w") as f:
                  raise ValueError()
                  f.write("Test")

    Instead of creating the file "test.txt" at the beginning and filling it with content
    later (which will never happen because of the exception thrown, which makes the file
    existing but the task actually not finished), the file will be written to a temporary
    file first and copied to its final location at the end of the run function (but only if there
    was no error).

    Warning:
        The decorator only edits the function :meth:`b2luigi.Task.get_output_file_name`. If you are using
        the output directly, you have to take care of using the temporary path correctly by yourself!

    """

    @wraps(run_function)
    def run(self):
        with TemporaryFileContextManager(self):
            run_function(self)

    return run



import b2luigi
import os


class MyTask(b2luigi.DispatchableTask):
    env = {"MY_SECOND_SECRET_VARIABLE": "47"}

    def process(self):
        print("MY_SECRET_VARIABLE", os.getenv("MY_SECRET_VARIABLE"))
        print("MY_SECOND_SECRET_VARIABLE", os.getenv("MY_SECOND_SECRET_VARIABLE"))


if __name__ == "__main__":
    # We cheat a bit and create the setup script here:
    with open("setup_script.sh", "w") as f:
        f.write("""echo "Setup Script called"\nexport MY_SECRET_VARIABLE=42""")

    b2luigi.set_setting("env_script", os.path.abspath("setup_script.sh"))
    b2luigi.process(MyTask())



import os
import stat
import subprocess

from b2luigi.core.settings import get_setting
from b2luigi.core.utils import (
    add_on_failure_function,
    create_cmd_from_task,
    get_filename,
    get_log_file_dir,
    get_task_file_dir,
    map_folder,
)


def create_executable_wrapper(task):
    """
    To incorporate all settings (environment, working paths, remote or locally)
    we create an executable bash script which is called instead of the application
    and which will setup everything accordingly before doing the actual work.
    """
    shell = get_setting("shell", task=task, default="bash")
    executable_wrapper_content = [f"#!/bin/{shell}", "set -e"]

    # 1. First part is the folder we need to change if given
    working_dir = get_setting("working_dir", task=task, default=os.path.abspath(os.path.dirname(get_filename())))
    executable_wrapper_content.append(f"cd {working_dir}")

    executable_wrapper_content.append("echo 'Working in the folder:'; pwd")

    # 2. Second part of the executable wrapper, the environment.
    executable_wrapper_content.append("echo 'Setting up the environment'")
    # (a) If given, use the environment script
    env_setup_script = get_setting("env_script", task=task, default="")
    if env_setup_script:
        # The script will be called from the directory of the script. So we have to make sure the
        # env_script is reachable from there (not from where we are currently)
        if not os.path.isfile(map_folder(env_setup_script)):
            raise FileNotFoundError(f"Environment setup script {env_setup_script} does not exist.")
        executable_wrapper_content.append(f"source {env_setup_script}")

    # (b) Now override with any environment from the task or settings
    env_overrides = get_setting("env", task=task, default={})
    for key, value in env_overrides.items():
        value = value.replace("'", "'''")
        value = f"'{value}'"
        executable_wrapper_content.append(f"export {key}={value}")

    executable_wrapper_content.append("echo 'Current environment:'; env")

    # 3. Third part is to call the actual program
    command = " ".join(create_cmd_from_task(task))
    executable_wrapper_content.append("echo 'Will now execute the program'")
    executable_wrapper_content.append(f"exec {command}")

    # Now we can write the file
    executable_file_dir = get_task_file_dir(task)
    os.makedirs(executable_file_dir, exist_ok=True)

    executable_wrapper_path = os.path.join(executable_file_dir, "executable_wrapper.sh")

    with open(executable_wrapper_path, "w") as f:
        f.write("\n".join(executable_wrapper_content))

    # make wrapper executable
    st = os.stat(executable_wrapper_path)
    os.chmod(executable_wrapper_path, st.st_mode | stat.S_IEXEC)

    return executable_wrapper_path


def run_task_remote(task):
    """
    Run a given task "remotely", which means
    create an exectable script and call it via a subprocess
    call.
    """
    log_file_dir = get_log_file_dir(task)
    os.makedirs(log_file_dir, exist_ok=True)
    stdout_log_file = os.path.join(log_file_dir, "stdout")
    stderr_log_file = os.path.join(log_file_dir, "stderr")

    executable_file = create_executable_wrapper(task)

    add_on_failure_function(task)

    with open(stdout_log_file, "w") as stdout_file:
        with open(stderr_log_file, "w") as stderr_file:
            return_code = subprocess.call([executable_file], stdout=stdout_file, stderr=stderr_file)

    if return_code:
        raise RuntimeError(f"Execution failed with return code {return_code}")



import os

from ..helpers import B2LuigiTestCase


class TemporaryWrapperTestCase(B2LuigiTestCase):
    def test_result_dir(self):
        self.call_file("core/folder_structures.py")

        self.assertTrue(os.path.exists("results/test.txt"))

        with open("results/test.txt", "r") as f:
            self.assertEqual(f.read(), "Test")

    def test_result_dir_from_other_location(self):
        os.makedirs("some_other_folder", exist_ok=True)
        self.call_file("core/folder_structures.py", cwd="some_other_folder")

        self.assertTrue(os.path.exists("results/test.txt"))
        self.assertTrue(os.path.exists("some_other_folder"))
        self.assertFalse(os.path.exists("some_other_folder/results/test.txt"))

        with open("results/test.txt", "r") as f:
            self.assertEqual(f.read(), "Test")



import os
from glob import glob

from ..helpers import B2LuigiTestCase


class ExampleTestCase(B2LuigiTestCase):
    def test_examples(self):
        for file_name in glob(os.path.join(os.path.dirname(__file__), "*.py")):
            short_file_name = os.path.basename(file_name)

            if short_file_name.startswith("test_") or short_file_name.startswith("__"):
                continue

            self.call_file(os.path.join("doc_examples/", os.path.basename(file_name)))



import json
import os
import contextlib
import warnings

# The global object hosting the current settings
_current_global_settings = {}


def get_setting(key, default=None, task=None, deprecated_keys=None):
    """
    ``b2luigi`` adds a settings management to ``luigi``
    and also uses it at various places.
    Many batch systems, the output and log path, the environment
    etc. is controlled via these settings.

    There are four ways settings could be defined.
    They are used in the following order (an earlier setting
    overrides a later one):

    1. If the currently processed (or scheduled) task has a property
       of the given name, it is used.
       Please note that you can either set the property directly, e.g.

       .. code-block:: python

         class MyTask(b2luigi.Task):
             batch_system = "htcondor"

       or by using a function (which might even depend on the parameters)

       .. code-block:: python

         class MyTask(b2luigi.Task):
             @property
             def batch_system(self):
                 return "htcondor"

       The latter is especially useful for batch system specific settings
       such as requested wall time etc.

    2. Settings set directly by the user in your script with a call to
       :meth:`b2luigi.set_setting`.
    3. Settings specified in the ``settings.json`` in the folder of your
       script *or any folder above that*.
       This makes it possible to have general project settings (e.g. the output path
       or the batch system) and a specific ``settings.json`` for your sub-project.

    With this function, you can get the current value of a specific setting with the given key.
    If there is no setting defined with this name,
    either the default is returned or, if you did not supply any default, a value error is raised.

    Settings can be of any type, but are mostly strings.

    Parameters:
        key (:obj:`str`): The name of the parameter to query.
        task: (:obj:`b2luigi.Task`): If given, check if the task has a parameter
            with this name.
        default (optional): If there is no setting which the name,
            either return this default or if it is not set,
            raise a ValueError.
        deprecated_keys (:obj:`List`): Former names of this setting,
            will throw a warning when still used
    """
    # First check if the correct name is set. If yes, just use it
    try:
        return _get_setting_implementation(key=key, task=task)
    except KeyError:
        pass

    # Ok, now test the deprecated setting names, but issue a warning
    if deprecated_keys:
        for deprecated_key in deprecated_keys:
            try:
                value = _get_setting_implementation(key=deprecated_key, task=task)
                _warn_deprecated_setting(deprecated_key, key)
                return value
            except KeyError:
                pass

    # Still not found? At this point we can only return the default or raise an error
    if default is None:
        raise ValueError(f"No settings found for {key}!")
    return default


def set_setting(key, value):
    """
    Set the setting with the specified name - overriding any ``setting.json``.
    If you want to have task specific settings, create a
    parameter with the given name or your task.
    """
    _current_global_settings[key] = value


def clear_setting(key):
    """
    Clear the setting with the given key
    """
    try:
        del _current_global_settings[key]
    except KeyError:
        pass


def _setting_file_iterator():
    path = os.getcwd()

    while True:
        json_file = os.path.join(path, "settings.json")
        if os.path.exists(json_file):
            yield json_file

        path = os.path.split(path)[0]
        if path == "/":
            break


@contextlib.contextmanager
def with_new_settings():
    global _current_global_settings
    old_settings = _current_global_settings.copy()
    _current_global_settings = {}

    yield

    _current_global_settings = old_settings.copy()


def _get_setting_implementation(key, task):
    """
    Implementation of the settings retrieval.
    Either get it from the task,
    or from the user-defined settings
    or from the setting files.
    If nothing is set, raise a KeyError.
    """
    # First check if the task has an attribute with this name
    if task:
        try:
            return getattr(task, key)
        except AttributeError:
            pass

    # Then check if the setting was set explicitely
    try:
        return _current_global_settings[key]
    except KeyError:
        pass

    # And finally check the settings files
    for settings_file in _setting_file_iterator():
        try:
            with open(settings_file, "r") as f:
                j = json.load(f)
                return j[key]
        except KeyError:
            pass

    # The setting was not found, so raise a KeyError
    raise KeyError(f"No settings found for {key}!")


class DeprecatedSettingsWarning(DeprecationWarning):
    pass


def _warn_deprecated_setting(setting_name, new_name):
    warnings.warn(
        f"The setting with the name {setting_name} is deprecated. "
        f"Please use {new_name} instead. Future versions might remove this setting.",
        DeprecatedSettingsWarning,
    )



import collections
import os
import shutil

import b2luigi
from b2luigi.basf2_helper.targets import ROOTLocalTarget
from b2luigi.basf2_helper.utils import get_basf2_git_hash

import subprocess

from b2luigi.core.utils import create_output_dirs, get_serialized_parameters


class Basf2Task(b2luigi.DispatchableTask):
    git_hash = b2luigi.Parameter(default=get_basf2_git_hash())

    def get_output_file_target(self, *args, **kwargs):
        file_name = self.get_output_file_name(*args, **kwargs)
        if os.path.splitext(file_name)[-1] == ".root":
            return ROOTLocalTarget(file_name)
        return super().get_output_file_target(*args, **kwargs)

    def get_serialized_parameters(self):
        serialized_parameters = get_serialized_parameters(self)

        # Git hash should go to the front
        return_dict = collections.OrderedDict()
        return_dict["git_hash"] = serialized_parameters["git_hash"]

        for key, value in serialized_parameters.items():
            return_dict[key] = value

        return return_dict


class Basf2PathTask(Basf2Task):
    num_processes = b2luigi.IntParameter(significant=False, default=0)
    max_event = b2luigi.IntParameter(significant=False, default=0)

    def create_path(self):
        raise NotImplementedError()

    @b2luigi.on_temporary_files
    def process(self):
        assert get_basf2_git_hash() == self.git_hash

        try:
            import basf2
        except ImportError:
            raise ImportError("Can not find basf2. Can not use the basf2 task.")

        if self.num_processes:
            basf2.set_nprocesses(self.num_processes)

        path = self.create_path()

        path.add_module("Progress")
        basf2.print_path(path)
        max_event = self.max_event if self.max_event else 0
        basf2.process(path=path, max_event=max_event)

        print(basf2.statistics)


class SimplifiedOutputBasf2Task(Basf2PathTask):
    def create_path(self):
        raise NotImplementedError()

    def output(self):
        path = self.create_path()
        outputs = []

        for module in path.modules():
            if module.type() == "RootOutput":
                for param in module.available_params():
                    if param.name == "outputFileName":
                        outputs.append(ROOTLocalTarget(param.values))

        return outputs


class MergerTask(Basf2Task):
    cmd = []

    def output(self):
        for key, _ in self.get_input_file_names().items():
            if hasattr(self, "keys") and key not in self.keys:
                continue

            yield self.add_to_output(key)

    @b2luigi.on_temporary_files
    def process(self):
        create_output_dirs(self)

        for key, file_list in self.get_input_file_names().items():
            if hasattr(self, "keys") and key not in self.keys:
                continue

            args = self.cmd + [self.get_output_file_name(key)] + file_list
            subprocess.check_call(args)


class HaddTask(MergerTask):
    cmd = ["hadd", "-f"]


class Basf2FileMergeTask(MergerTask):
    cmd = ["b2file-merge", "-f"]


class Basf2nTupleMergeTask(MergerTask):
    @property
    def cmd(self):
        "Command to use to merge basf2 tuple files."
        # ``fei_merge_files`` has been renamed to ``analysis-fei-mergefiles``, use
        # the newer command if it exists in the release.
        new_cmd_name = "analysis-fei-mergefiles"
        old_cmd_name = "fei_merge_files"
        if shutil.which(new_cmd_name):
            return [new_cmd_name]
        return [old_cmd_name]



from ..helpers import B2LuigiTestCase

import b2luigi
from b2luigi.core.utils import get_filled_params


class TaskTestCase(B2LuigiTestCase):
    def test_file_path_usage(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("file_a")
                yield self.add_to_output("file_b")

        task = TaskA(some_parameter=3)

        b2luigi.set_setting("result_dir", "results/some_crazy_path")

        self.assertEqual(get_filled_params(task), {"some_parameter": 3})
        self.assertFalse(task.get_input_file_names())
        self.assertRaises(KeyError, lambda: task._get_input_targets("some_file"))
        self.assertEqual(task._get_output_target("file_a").path, task.get_output_file_name("file_a"))
        self.assertIn("file_a", task.get_output_file_name("file_a"))
        self.assertIn("file_b", task.get_output_file_name("file_b"))
        self.assertIn("some_parameter=3", task.get_output_file_name("file_a"))
        self.assertIn("some_crazy_path", task.get_output_file_name("file_a"))

    def test_dependencies(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("file_a")

        @b2luigi.requires(TaskA)
        class TaskB(b2luigi.Task):
            def output(self):
                yield self.add_to_output("file_b")

        task = TaskB(some_parameter=42)

        self.assertEqual(get_filled_params(task), {"some_parameter": 42})
        self.assertEqual(len(task._get_input_targets("file_a")), 1)
        self.assertEqual(len(task.get_input_file_names("file_a")), 1)
        self.assertEqual(len(task.get_input_file_names().keys()), 1)
        self.assertEqual(task._get_input_targets("file_a")[0].path, task.get_input_file_names("file_a")[0])
        self.assertEqual(task._get_output_target("file_b").path, task.get_output_file_name("file_b"))
        self.assertIn("file_b", task.get_output_file_name("file_b"))
        self.assertIn("some_parameter=42", task.get_output_file_name("file_b"))
        self.assertEqual(len(list(task.get_all_input_file_names())), 1)
        self.assertEqual(len(list(task.get_all_output_file_names())), 1)
        self.assertIn("some_parameter=42/file_a", list(task.get_all_input_file_names())[0])
        self.assertIn("some_parameter=42/file_b", list(task.get_all_output_file_names())[0])

    def test_many_dependencies(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("file_a")

        class TaskB(b2luigi.Task):
            def requires(self):
                for i in range(100):
                    yield self.clone(TaskA, some_parameter=i)

        task = TaskB()

        self.assertEqual(len(task._get_input_targets("file_a")), 100)
        self.assertEqual(len(task.get_input_file_names("file_a")), 100)
        self.assertEqual(len(task.get_input_file_names().keys()), 1)

        self.assertEqual(len(task.get_input_file_names()["file_a"]), 100)

        input_file_names = task.get_input_file_names("file_a")

        # We are only interested in the last 3 parts of the folder
        input_file_names = ["/".join(x.split("/")[-3:]) for x in input_file_names]

        for i in range(100):
            self.assertIn(f"results/some_parameter={i}/file_a", input_file_names)



import b2luigi


class AlwaysExistingTarget(b2luigi.Target):
    """
    Target that always exists
    """

    def exists(self):
        return True


class DoNothingTask(b2luigi.ExternalTask):
    """
    Dummy task that is always already complete and needs nothing to be done.

    Useful for testing ``b2luigi.process`` just for its CLI argument parsing
    with some dummy task.
    """

    def output(self):
        return AlwaysExistingTarget()


if __name__ == "__main__":
    b2luigi.process(DoNothingTask(), ignore_additional_command_line_args=False)
